using System;
using System.Collections.Generic;
using System.Linq;
using System.Text.RegularExpressions;
using TMPro;
using UnityEngine;
#if UNITY_LOCALIZATION
using UnityEditor.Localization;
using UnityEngine.Localization.SmartFormat;
using UnityEngine.Localization.SmartFormat.Core.Extensions;
using UnityEngine.Localization.SmartFormat.Core.Formatting;
using UnityEngine.Localization.SmartFormat.Core.Parsing;
using UnityEngine.Localization.Components;
using UnityEngine.Localization.Settings;
using UnityEngine.Localization.Tables;
using Object = UnityEngine.Object;
#endif

namespace Magify
{
    public static class BundleFontCustomizer
    {
        private const string AlwaysIncludedSymbols
            = "0123456789" // All numbers
            + "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz" // latin symbols
            + ".,/₴$€£¥₪₨₩₦₫ƒрč₭₮₱₡₲₵ł" // default symbols
            + " \n\r_"; // special numbers
            // + "₾₼₺₸₹៛դր֏৳₽฿₿" // advanced symbols

        private static readonly HashSet<TMP_FontAsset> _allFontAssets = new();
        private static readonly Dictionary<TMP_Text, HashSet<TMP_FontAsset>> _textToFontMap = new();
        private static GameObject _activeCreative;

        public static void PrepareFonts(GameObject creativePrefab)
        {
            _activeCreative = creativePrefab;
            var tmpTexts = creativePrefab.GetComponentsInChildren<TMP_Text>(true);

            foreach (var tmpText in tmpTexts)
            {
                var fontAssets = CollectFonts(tmpText);
                StoreToAllFontAssets(fontAssets);
                _textToFontMap.Add(tmpText, fontAssets);
            }

            ClearFonts();
            PrepareNeededGlyphs();
            SetFontSettingsToBuild();
        }

        public static void RevertFontsSettings()
        {
            ClearFonts();
            _allFontAssets.Clear();
            _textToFontMap.Clear();
            _activeCreative = null;
        }

        public static void RemoveLocalizationScripts(GameObject editingScopePrefabContentsRoot)
        {
            var localizeStringEvents = editingScopePrefabContentsRoot.GetComponentsInChildren<LocalizeStringEvent>(true);
            foreach (var localizeStringEvent in localizeStringEvents)
            {
                Object.DestroyImmediate(localizeStringEvent, true);
            }
        }

        private static void SetFontSettingsToBuild()
        {
            foreach (var fontAsset in _allFontAssets)
            {
                fontAsset.atlasPopulationMode = AtlasPopulationMode.Static;
            }
        }

        private static void ClearFonts()
        {
            foreach (var fontAsset in _allFontAssets)
            {
                fontAsset.atlasPopulationMode = AtlasPopulationMode.Dynamic;
                fontAsset.ClearFontAssetData(true);
            }
        }

        private static void StoreToAllFontAssets(HashSet<TMP_FontAsset> tmpFontAssets)
        {
            foreach (var fontAsset in tmpFontAssets)
            {
                _allFontAssets.Add(fontAsset);
            }
        }

        private static HashSet<TMP_FontAsset> CollectFonts(TMP_Text tmpText)
        {
            var result = new HashSet<TMP_FontAsset>();

            if (tmpText.font == null)
            {
                Debug.LogError("Text with empty font has been found - can't continue work", tmpText);
                return result;
            }

            result.Add(tmpText.font);
            collectFallbackFonts(tmpText.font, result);

            return result;

            void collectFallbackFonts(TMP_FontAsset fontAsset, HashSet<TMP_FontAsset> fontsLookup)
            {
                foreach (var fallbackFontAsset in fontAsset.fallbackFontAssetTable.Where(c => c != null))
                {
                    if (fontsLookup.Contains(fallbackFontAsset))
                    {
                        continue;
                    }

                    fontsLookup.Add(fallbackFontAsset);
                    collectFallbackFonts(fallbackFontAsset, fontsLookup);
                }
            }
        }

        private static void PrepareNeededGlyphs()
        {
            var isUpperOrLowerCase = false;

            if (_allFontAssets.Any())
            {
                AddCharactersToFonts(_allFontAssets, AlwaysIncludedSymbols);
            }

            foreach (var pair in _textToFontMap)
            {
                var tmp = pair.Key;
                var fonts = pair.Value;
                LocalizedText(tmp);
                var usedGlyphs = GetCharacterForPreparing(tmp);
                usedGlyphs = RemoveHtmlTags(usedGlyphs);

                var fontStyle = tmp.fontStyle;
                if ((fontStyle & FontStyles.LowerCase) != 0)
                {
                    usedGlyphs = usedGlyphs.ToLower();
                    isUpperOrLowerCase = true;
                }
                else if ((fontStyle & (FontStyles.UpperCase | FontStyles.SmallCaps)) != 0)
                {
                    usedGlyphs = usedGlyphs.ToUpper();
                    isUpperOrLowerCase = true;
                }

                AddCharactersToFonts(fonts, usedGlyphs);
            }

            CheckLocalizationCache(isUpperOrLowerCase);
        }

        private static void CheckLocalizationCache(bool isUpperOrLowerCase)
        {
#if UNITY_LOCALIZATION
            foreach (var pair in _textToFontMap)
            {
                var tmp = pair.Key;

                var bundleLocalization = tmp.GetComponent<BundleLocalizationCache>();
                if (bundleLocalization != null)
                {
                    foreach (var key in bundleLocalization.LocalizeReferences)
                    {
                        var localizedText = GetUsedLocalizedCharacters(LocalizationSettings.SelectedLocale.Identifier.Code, key.TableEntryReference);
                        if (isUpperOrLowerCase)
                        {
                            localizedText = localizedText.ToLower() + localizedText.ToUpper();
                        }
                        AddCharactersToFonts(_allFontAssets, localizedText);
                    }

                    Object.DestroyImmediate(bundleLocalization, true);
                }
            }
#endif
        }

        private static void AddCharactersToFonts(HashSet<TMP_FontAsset> fonts, string characters)
        {
            foreach (var fontAsset in fonts)
            {
                if (fontAsset.HasCharacters(characters, out var missingCharactersList))
                {
                    characters = string.Empty;
                    break;
                }

                fontAsset.TryAddCharacters(string.Join("", missingCharactersList), out var missingCharacters);
                characters = missingCharacters;
                if (missingCharacters.Length == 0)
                {
                    break;
                }
            }

            if (characters.Length > 0)
            {
#if UNITY_LOCALIZATION
                var currentLocale = LocalizationSettings.SelectedLocale;
#else
                var currentLocale = "English";
#endif
                throw new Exception($"Can't build creative <color=yellow>{_activeCreative.name}</color> " +
                                    $"for localization <color=yellow>{currentLocale}</color>. " +
                                    "$Because can't find in all fonts</color>" +
                                    $" this symbols: <color=white>{string.Join(", ", characters)}</color>");
            }
        }

        private static void LocalizedText(TMP_Text tmp)
        {
#if UNITY_LOCALIZATION
            var localization = tmp.GetComponent<LocalizeStringEvent>();
            if (localization != null)
            {
                tmp.text = localization.StringReference.GetLocalizedString();
            }
#endif
        }

        private static string GetCharacterForPreparing(TMP_Text tmp)
        {
            var symbols = tmp.text;
#if UNITY_LOCALIZATION
            var localization = tmp.GetComponent<LocalizeStringEvent>();
            if (localization != null)
            {
                symbols += GetUsedLocalizedCharacters(LocalizationSettings.SelectedLocale.Identifier.Code, localization.StringReference.TableEntryReference);
            }
#endif
            return symbols;
        }

#if UNITY_LOCALIZATION
        internal static string GetUsedLocalizedCharacters(string cultureCode, TableEntryReference tableEntryReference)
        {
            var localizationCollection = LocalizationEditorSettings.GetStringTableCollection(UnityLocalizerTables.GlobalLocalizationTableName);
            var currentCultureTable = (StringTable)localizationCollection.GetTable(cultureCode);
            var localizeEntry = currentCultureTable.GetEntryFromReference(tableEntryReference);
            var smart = new SmartFormatterLiteralCharacterExtractor(LocalizationSettings.StringDatabase?.SmartFormatter);
            if (localizeEntry.IsSmart)
            {
                var literals = smart.ExtractLiteralsCharacters(localizeEntry.LocalizedValue);
                var sorted = literals.Distinct().OrderBy(c => c);
                return string.Concat(sorted);
            }

            return localizeEntry.LocalizedValue;
        }
#endif

        private static string RemoveHtmlTags(string str)
        {
            var rich = new Regex(@"<[^>]*>");

            if (rich.IsMatch(str))
            {
                str = rich.Replace(str, string.Empty);
            }

            return str;
        }
    }

#if UNITY_LOCALIZATION
    public class SmartFormatterLiteralCharacterExtractor : SmartFormatter
    {
        private IEnumerable<char> m_Characters;

        public SmartFormatterLiteralCharacterExtractor(SmartFormatter parent)
        {
            Settings = parent.Settings;
            Parser = parent.Parser;
            SourceExtensions.AddRange(parent.SourceExtensions);
            FormatterExtensions.AddRange(parent.FormatterExtensions);
        }

        public IEnumerable<char> ExtractLiteralsCharacters(string value)
        {
            m_Characters = "";
            Format(value, null);
            return m_Characters;
        }

        public override void Format(FormattingInfo formattingInfo)
        {
            foreach (var item in formattingInfo.Format.Items)
            {
                if (item is LiteralText literalItem)
                {
                    m_Characters = m_Characters.Concat(item.ToEnumerable());
                    continue;
                }

                // Otherwise, the item must be a placeholder.
                var placeholder = (Placeholder) item;
                var childFormattingInfo = formattingInfo.CreateChild(placeholder);

                var formatterName = childFormattingInfo.Placeholder.FormatterName;

                // Evaluate the named formatter (or, evaluate all "" formatters)
                foreach (var formatterExtension in FormatterExtensions)
                {
                    if (formatterExtension is IFormatterLiteralExtractor literalExtractor &&
                        formatterExtension.Names.Contains(formatterName))
                    {
                        literalExtractor.WriteAllLiterals(childFormattingInfo);
                    }
                }
            }
        }
    }
#endif
}