﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Text;
using System.Threading;
using Cysharp.Threading.Tasks;
using JetBrains.Annotations;
using Magify.Model;
using Newtonsoft.Json;
using UniRx;
using UnityEngine;

namespace Magify
{
    internal class AppStateManager : IInitializable, IDisposable
    {
        [NotNull]
        private static readonly MagifyLogger _logger = MagifyLogger.Get(LoggingScope.AppState);

        [NotNull]
        private readonly PlatformAPI _platform;
        [NotNull]
        private readonly ServerApi _serverApi;
        [NotNull]
        private readonly GeneralPrefs _generalPrefs;
        [NotNull]
        private readonly ClientIdProvider _clientIdProvider;
        [NotNull]
        private readonly CountersStorage _counters;
        [NotNull]
        private readonly AppStatePrefs _appStatePrefs;
        [NotNull]
        private readonly ClientStateConfig _clientStateConfig;
        [NotNull]
        private readonly AppStateBuilder _appStateBuilder;
        [NotNull]
        private readonly CompositeDisposable _disposables = new();
        [NotNull]
        private readonly Subject<SyncStateResult> _restoreStateCompletedSubject = new();

        private CancellationTokenSource _saveStateCts;
        private CancellationTokenSource _restoreStateCts;
        private CancellationTokenSource _authorizeUserCts;
        private bool _restoreStateCompletedCallbackInProgress;
        private bool _isInitialRestoreInProgress;

        [NotNull]
        public IObservable<SyncStateResult> OnRestoreStateCompleted => _restoreStateCompletedSubject;
        [NotNull]
        public IReactiveProperty<bool> SyncStateEnabled => _generalPrefs.SyncStateEnabled;

        public AppStateManager(
            [NotNull] PlatformAPI platform,
            [NotNull] ServerApi serverApi,
            [NotNull] GeneralPrefs generalPrefs,
            [NotNull] ClientIdProvider clientIdProvider,
            [NotNull] CountersStorage counters,
            [NotNull] AppStatePrefs appStatePrefs,
            [NotNull] ClientStateConfig clientStateConfig,
            [NotNull] string storagePath)
        {
            _platform = platform;
            _serverApi = serverApi;
            _generalPrefs = generalPrefs;
            _clientIdProvider = clientIdProvider;
            _counters = counters;
            _appStatePrefs = appStatePrefs;
            _clientStateConfig = clientStateConfig;
            _appStateBuilder = new AppStateBuilder(generalPrefs, counters, clientStateConfig, storagePath);
            SyncStateEnabled
                .Subscribe(enabled => _logger.Log($"Sync state {(enabled ? "enabled" : "disabled")}"))
                .AddTo(_disposables);
        }

        async void IInitializable.Initialize()
        {
            _clientIdProvider.ClientId
                .Where(token => SyncStateEnabled.Value && !string.IsNullOrWhiteSpace(token))
                .Subscribe(ClientIdChangedHandler)
                .AddTo(_disposables);

            await _clientIdProvider.WaitForLoadingFromCloudFinished();
            if (string.IsNullOrEmpty(_generalPrefs.AuthorizationToken.Value))
            {
                _ = _serverApi.GetAuthorizationTokenAsync(CancellationToken.None);
            }
            if (_appStatePrefs.IsPendingRestore.Value)
            {
                LaunchInitialStateRestore(_disposables.GetOrCreateToken()).Forget();
            }
        }

        void IDisposable.Dispose()
        {
            _disposables.Dispose();
            _saveStateCts?.Cancel();
            _saveStateCts?.Dispose();
            _restoreStateCts?.Cancel();
            _restoreStateCts?.Dispose();
        }

        public async UniTask<bool> RequestAuthTokenFor([NotNull] string provider, [NotNull] string token, CancellationToken cancellationToken)
        {
            _appStatePrefs.UpdateExternalAuthToken(provider, token);
            _serverApi.ResetAuthorizationToken();
            try
            {
                var authToken = await _serverApi.GetAuthorizationTokenAsync(LinkCancellationToken(cancellationToken).Token);
                if (authToken != null) _logger.Log($"Authorization token updated with external user token. Provider: {provider}; Token: {token}; Auth token: {authToken}");
                else _logger.LogError($"Failed to update authorization token with external user token. Provider: {provider}; Token: {token}");
                return authToken != null;
            }
            catch (MagifyAuthTokenLoadingCancelledException e)
            {
                _logger.LogException(e);
                return false;
            }
        }

        public async UniTask<SyncStateResult> SaveState(int? weight, CancellationToken cancellationToken)
        {
            _logger.Log($"Start saving application state. Weight: {weight?.ToString() ?? "null"}");

            await  _clientIdProvider.WaitForLoadingFromCloudFinished();
            if (cancellationToken.IsCancellationRequested)
            {
                return SyncStateResult.SaveCancelled(cancellationToken, weight);
            }

            SyncStateResult result;
            if (!SyncStateEnabled.Value)
            {
                result = SyncStateResult.SyncIsDisabled();
                _logger.LogError($"Can't save state. {result.ErrorMessage}");
                return result;
            }
            if (_appStatePrefs.IsPendingRestore.Value)
            {
                result = SyncStateResult.RestoreInProgress();
                _logger.LogError($"Can't save state. {result.ErrorMessage}");
                return result;
            }

            try
            {
                cancellationToken = LinkCancellationToken(cancellationToken).Token;
                var authResult = await AuthorizeUserIfNeeded(cancellationToken);
                if (authResult!.IsSuccess)
                {
                    _saveStateCts?.Cancel();
                    _saveStateCts?.Dispose();
                    _saveStateCts = CancellationTokenSource.CreateLinkedTokenSource(cancellationToken);
                    var isSaved = await _serverApi.SaveState(await _appStateBuilder.Build(cancellationToken, _platform.RuntimePlatform), weight, _saveStateCts!.Token);
                    if (isSaved)
                    {
                        _logger.Log($"Application state successfully saved. Weight: {weight?.ToString() ?? "null"}");
                        return SyncStateResult.Ok(weight);
                    }
                }

                var error = $"weight: {weight?.ToString() ?? "null"} " + ErrIfCancelled(cancellationToken);
                result = SyncStateResult.FailedToSave(error, weight);
                _logger.LogError(result.ErrorMessage);
                return result;
            }
            catch (OperationCanceledException e)
            {
                return SyncStateResult.SaveCancelled(e.CancellationToken, weight);
            }
        }

        public async UniTask<SyncStateResult> RestoreState(int? weight, CancellationToken cancellationToken, bool forceSync = false)
        {
            _logger.Log($"Start restoring application state. Weight: {weight?.ToString() ?? "null"}");

            await _clientIdProvider.WaitForLoadingFromCloudFinished();
            if (cancellationToken.IsCancellationRequested)
            {
                return SyncStateResult.SaveCancelled(cancellationToken, weight);
            }

            SyncStateResult result;
            if (_restoreStateCompletedCallbackInProgress)
            {
                result = SyncStateResult.RestoreCallbackInProgress(weight);
                _logger.LogError($"Can't restore state. {result.ErrorMessage}");
                return result;
            }
            if (!SyncStateEnabled.Value && !forceSync)
            {
                result = SyncStateResult.SyncIsDisabled();
                _logger.LogError(result.ErrorMessage);
                return result;
            }
            if (!forceSync && _appStatePrefs.IsPendingRestore.Value && weight != null)
            {
                result = SyncStateResult.RestoreInProgress();
                _logger.LogError($"Can't restore state. {result.ErrorMessage}");
                return result;
            }

            try
            {
                _appStatePrefs.IsRestoreInProgress.Value = true;
                cancellationToken = LinkCancellationToken(cancellationToken).Token;
                var authResult = await AuthorizeUserIfNeeded(cancellationToken);
                if (authResult!.IsSuccess)
                {
                    _restoreStateCts?.Cancel();
                    _restoreStateCts?.Dispose();
                    _restoreStateCts = CancellationTokenSource.CreateLinkedTokenSource(cancellationToken);
                    var applicationState = await _serverApi.RestoreState(weight, _restoreStateCts!.Token);
                    var isEmpty = applicationState == null;
                    var internalStateRestored = false;
                    if (!isEmpty)
                    {
                        _logger.Log($"Application state successfully loaded. Weight: {weight?.ToString() ?? "null"}");
                        var parsedAppState = _appStateBuilder.Parse(applicationState, _platform.RuntimePlatform);
                        using (new DebugAppStateRestoreScope())
                        {
                            ApplyApplicationState(parsedAppState, cancellationToken, out internalStateRestored);
                        }
                        _logger.Log($"Application state successfully restored. Weight: {weight?.ToString() ?? "null"}");
                    }
                    else
                    {
                        _logger.Log("Restores application state is null, skipping writing to files.");
                    }
                    _appStatePrefs.IsPendingRestore.Value = false;
                    _appStatePrefs.IsRestoreInProgress.Value = false;
                    _restoreStateCompletedCallbackInProgress = true;
                    result = isEmpty ? SyncStateResult.NothingToRestore(weight) : SyncStateResult.Ok(weight, internalStateRestored);
                    _restoreStateCompletedSubject.OnNext(result);
                    _restoreStateCompletedCallbackInProgress = false;
                    return result;
                }

                result = SyncStateResult.FailedToRestore($"weight: {weight?.ToString() ?? "null"} , content: null" + ErrIfCancelled(cancellationToken), weight);
                _logger.LogError(result.ErrorMessage);
                return result;
            }
            catch (OperationCanceledException e)
            {
                return SyncStateResult.RestoreCancelled(e.CancellationToken, weight);
            }
            catch (MagifyFailedToParseRestoreProgressResponseException e)
            {
                return SyncStateResult.FailedToRestore(e.Message, weight);
            }
            finally
            {
                _restoreStateCompletedCallbackInProgress = false;
                _appStatePrefs.IsRestoreInProgress.Value = false;
            }
        }

        private void ApplyApplicationState([CanBeNull] ApplicationState state, CancellationToken cancellationToken, out bool internalStateRestored)
        {
            _logger.Log($"Applying application state: \n{JsonFacade.SerializeObject(state, Formatting.Indented)}");
            if (state?.ClientState != null && _clientStateConfig.AppStateProvider != null)
            {
                try
                {
                    _clientStateConfig.AppStateProvider.HandleState(state.ClientState, cancellationToken);
                }
                catch (Exception e)
                {
                    var exception = new ProviderFailedToHandleClientStateException(_clientStateConfig.AppStateProvider, e);
                    _logger.LogException(exception);
                    throw exception;
                }
            }
            var sdk = state?.Sdk(_platform.RuntimePlatform);
            internalStateRestored = sdk is { IsEmpty: false };
            if (sdk?.GeneralPrefs != null)
            {
                _logger.Log("Applying general prefs");
                _generalPrefs.RewriteContent(sdk.GeneralPrefs);
            }
            if (sdk?.CountersPrefs != null)
            {
                _logger.Log("Applying counters prefs");
                _counters.RewriteContent(sdk.CountersPrefs);
            }
        }

        private async UniTask<SyncStateResult> AuthorizeUserIfNeeded(CancellationToken cancellationToken)
        {
            if (_appStatePrefs.IsFbAuthorizationCompleted.Value)
            {
                return SyncStateResult.Ok(null);
            }

            _authorizeUserCts?.Cancel();
            _authorizeUserCts?.Dispose();
            _authorizeUserCts = CancellationTokenSource.CreateLinkedTokenSource(cancellationToken);
            cancellationToken = _authorizeUserCts.Token;
            _logger.Log($"Authorization is required. Client id: {_clientIdProvider.ClientId.Value}");
            var userId = await _serverApi.AuthorizeUser(_clientIdProvider.ClientId.Value, cancellationToken);
            if (userId == null)
            {
                var result = SyncStateResult.FailedToAuthorize(_clientIdProvider.ClientId.Value + ErrIfCancelled(cancellationToken));
                _logger.LogError(result.ErrorMessage);
                return result;
            }
            _appStatePrefs.IsFbAuthorizationCompleted.Value = true;
            _logger.Log($"Authorization successfully completed. User id: {userId}");
            return SyncStateResult.Ok(null);
        }

        private void ClientIdChangedHandler([NotNull] string authToken)
        {
            if (_appStatePrefs.IsPendingRestore.Value)
            {
                LaunchInitialStateRestore(_disposables.GetOrCreateToken()).Forget();
            }
        }

        private async UniTaskVoid LaunchInitialStateRestore(CancellationToken cancellationToken)
        {
            _logger.Log($"{nameof(LaunchInitialStateRestore)} called");
            if (_isInitialRestoreInProgress)
            {
                _logger.Log("Initial state restore already ran.");
                return;
            }
            try
            {
                _isInitialRestoreInProgress = true;
                bool continueRestore;
                var backoff = new ExponentialBackoff();
                do
                {
                    var result = await RestoreState(null, CancellationToken.None, forceSync: true)!;
                    continueRestore = result?.Kind is not (SyncStateResult.ResultKind.Success or SyncStateResult.ResultKind.NothingToRestore);
                    if (continueRestore)
                    {
                        _logger.LogWarning($"Failed to restore initial state. {result?.ErrorMessage}");
                        await UniTask.Delay(Convert.ToInt32(backoff.NextDelay()), DelayType.Realtime, cancellationToken: cancellationToken);
                    }

                } while (continueRestore);
            }
            finally
            {
                _isInitialRestoreInProgress = false;
            }
        }

        [NotNull]
        private CancellationTokenSource LinkCancellationToken(CancellationToken cancellationToken)
        {
            return CancellationTokenSource.CreateLinkedTokenSource(cancellationToken, _disposables.GetOrCreateToken());
        }

        [NotNull]
        private static string ErrIfCancelled(CancellationToken cancellationToken)
        {
            return cancellationToken.IsCancellationRequested ? "(request cancelled)" : string.Empty;
        }

        private class DebugAppStateRestoreScope : IDisposable
        {
#if MAGIFY_VERBOSE_LOGGING
            [CanBeNull]
            private readonly IDisposable _disposable;
            [NotNull]
            private readonly Dictionary<string, string> _map = new();
#endif

#if MAGIFY_VERBOSE_LOGGING
            public DebugAppStateRestoreScope()
            {
                _disposable = BinaryStorage.OnStorageSavedAsJson
                    .Subscribe(tuple => _map.TryAdd(tuple.Name, tuple.Json));
            }
#endif

            public void Dispose()
            {
#if MAGIFY_VERBOSE_LOGGING
                _disposable?.Dispose();
                var builder = new StringBuilder();
                builder.Append("Prettified restored progress:");
                foreach (var (key, value) in _map)
                {
                    builder
                        .Append('\n')
                        .Append(key)
                        .Append(':')
                        .Append('\n')
                        .Append(value)
                        .Append('\n')
                        .Append("--------------------");
                }
                var str = builder.ToString();
                _logger.Log(str);
                var path = Path.Combine(PackageInfo.PersistentPath, "pretty-progress.txt");
                if (!Directory.Exists(Path.GetDirectoryName(path)))
                    Directory.CreateDirectory(Path.GetDirectoryName(path));
                File.WriteAllText(path, str);
#endif
            }
        }
/*
        private class RestoreProgressScopeSource
        {
            [NotNull]
            private readonly AppStatePrefs _appStatePrefs;
            [NotNull]
            private readonly object _lock = new();
            private int _counter;

            public RestoreProgressScopeSource([NotNull] AppStatePrefs appStatePrefs)
            {
                _appStatePrefs = appStatePrefs;
            }

            [NotNull]
            public Token Get()
            {
                lock (_lock)
                {
                    if (_counter == 0)
                    {
                        _appStatePrefs.IsUserRestoreInProgress.Value = true;
                    }
                    var token = new Token(OnTokenDisposed);
                    _counter++;
                    return token;
                }
            }

            private void OnTokenDisposed()
            {
                lock (_lock)
                {
                    _counter--;
                    if (_counter == 0)
                    {
                        _appStatePrefs.IsUserRestoreInProgress.Value = false;
                    }
                }
            }

            public class Token : IDisposable
            {
                [NotNull]
                private readonly object _lock = new();
                private readonly Action _onCurrentDisposed;
                private bool _disposed;

                public Token([NotNull] Action onCurrentDisposed)
                {
                    _onCurrentDisposed = onCurrentDisposed;
                }

                public void Dispose()
                {
                    lock (_lock)
                    {
                        if (!_disposed)
                        {
                            _disposed = true;
                            _onCurrentDisposed?.Invoke();
                        }
                    }
                }
            }
        }
        */
    }
}