using System;
using System.Collections.Generic;
using JetBrains.Annotations;
using Magify.Types;
using UniRx;
using UnityEngine;
using UnityEngine.Pool;

namespace Magify
{
    internal class GeneralPrefs : ILegacyMigrator, IInitializable, IDisposable
    {
        private const string KeyMigratedUser = "migrated_user";

        internal const string KeyClientId = "client_id";
        private const string KeyFirstInstalledVersion = "first_version";
        private const string KeyFirstLaunchDate = "first_launch";

        private const string KeySessionCounter = "session_counter";
        private const string KeyVersionSessionCounter = "version_session_counter";
        private const string KeySessionId = "session_id";
        private const string KeySessionOpenTime = "session_open_time";
        private const string KeySessionLastTime = "session_last_time";

        private const string KeyAuthToken = "auth_token";
        private const string KeyAuthTokenExpired = "auth_token_expired";

        private const string KeyLanguageCode = "language_code";
        private const string KeyCountryCode = "country_code";

        private const string KeySubscriptionState = "subscription_state";
        private const string KeyInAppStatus = "in_app_status";
        private const string KeyAuthorizationStatus = "authorization_status";
        private const string KeyReferrerId = "referrer_id";
        private const string KeyAdjustId = "adjust_id";
        private const string KeyGpsAdId = "gps_ad_id";
        private const string KeyGdprStatus = "gdpr_granted";
        private const string KeyAttAuthorizationStatus = "att_authorization_status";
        private const string KeySyncStateEnabled = "sync_state_enabled";

        private const string KeySourceNetwork = "source_network";
        private const string KeySourceCampaign = "source_campaign";
        private const string KeySourceGroup = "source_group";

        private const string KeyDefaultCampaignTypes = "default_campaign_types";
        private const string KeyProductsReceiptTime = "products_by_dates";

        private const string KeyPurchasedInApps = "purch_products";
        private const string KeyPurchasedSubscriptions = "purch_subs_products";
        private const string KeyUsedFreeProducts = "free_products";

        private const string KeyFirebaseInstanceID = "firebase_instance_id";

        private const string KeyNextDailyResetTime = "next_daily_reset_time";

        private const string KeyCampaignsContextScopes = "CampaignsContextScopes";

        internal const SubscriptionStatus DefaultSubscriptionStatus = Magify.SubscriptionStatus.Inactive;
        internal const InAppStatus DefaultInAppStatus = Magify.InAppStatus.NeverPurchased;
        internal const AuthorizationStatus DefaultAuthorizationStatus = Magify.AuthorizationStatus.Unknown;

        [NotNull]
        private readonly BinaryStorage _storage;
        [NotNull]
        private readonly BinaryStorage _localSubStorage;
        [NotNull]
        private readonly PlatformAPI _platform;
        [NotNull]
        private readonly IDictionary<string, int> _versionSessionCounter;
        [NotNull]
        private readonly CompositeDisposable _disposable = GenericPool<CompositeDisposable>.Get()!;

        [NotNull]
        public IReactiveProperty<bool> IsFirstAppLaunchTracked { get; } = new ReactiveProperty<bool>(false);

        public bool MigratedUser
        {
            get => _storage.Get<bool>(KeyMigratedUser);
            private set => _storage.Set(KeyMigratedUser, value);
        }

        [NotNull]
        public NotNullReactiveProperty<string> ClientId { get; }

        public string FirstInstalledVersion
        {
            get => _storage.Get<string>(KeyFirstInstalledVersion);
            set => _storage.Set(KeyFirstInstalledVersion, value);
        }

        public DateTime FirstLaunchDate
        {
            get => _storage.Get(KeyFirstLaunchDate, default(DateTime));
            set => _storage.Set(KeyFirstLaunchDate, value);
        }

        public bool IsNewUser => FirstInstalledVersion == Application.version;
        public IReactiveProperty<bool> AuthorizationTokenExpired { get; }
        public IReactiveProperty<string> AuthorizationToken { get; }
        public IReactiveProperty<string> LanguageCode { get; }
        public IReactiveProperty<string> CountryCode { get; }
        public IReactiveProperty<SubscriptionStatus> SubscriptionStatus { get; }
        public IReactiveProperty<InAppStatus> InAppStatus { get; }
        public IReactiveProperty<AuthorizationStatus> AuthorizationStatus { get; }
        public IReactiveProperty<string> ReferrerId { get; }
        public IReactiveProperty<string> AdjustId { get; }
        public IReactiveProperty<string> GpsAdId { get; }
        public IReactiveProperty<bool> IsGdprGranted { get; }
        public IReactiveProperty<bool?> AttAuthorizationStatus { get; }
        public ISet<string> UsedDefaultCampaignTypes { get; }
        public IDictionary<string, DateTime> ProductsReceiptTimestamps { get; }
        public IReactiveProperty<bool> SyncStateEnabled { get; }

        public ISet<string> PurchasedInAppProducts { get; }
        public ISet<string> PurchasedSubscriptionProducts { get; }
        public ISet<string> UsedFreeProducts { get; }
        public ICollection<string> AllUsedProducts { get; }

        public IReactiveProperty<string> FirebaseInstanceId { get; }

        public IReactiveProperty<int> GlobalSessionCounter { get; }

        public IReactiveProperty<string> SessionId { get; }

        [NotNull]
        public IReactiveProperty<ConfigScope> RemoteContextScopes { get; }

        public int VersionSessionCounter
        {
            get => _versionSessionCounter.TryGetValue(Application.version, out var value) ? value : 0;
            set => _versionSessionCounter[Application.version] = value;
        }

        public long SessionStarted
        {
            get => _storage.Get(KeySessionOpenTime, 0L);
            set => _storage.Set(KeySessionOpenTime, value);
        }

        public long LastActivity
        {
            get => _storage.Get(KeySessionLastTime, 0L);
            set => _storage.Set(KeySessionLastTime, value);
        }

        public MediaSource MediaSource
        {
            get
            {
                var network = _localSubStorage.Get<string>(KeySourceNetwork);
                var campaign = _localSubStorage.Get<string>(KeySourceCampaign);
                var group = _localSubStorage.Get<string>(KeySourceGroup);
                return new MediaSource(network, campaign, group);
            }
            set
            {
                using var _ = _localSubStorage.MultipleChangeScope();
                _localSubStorage.Set(KeySourceNetwork, value.NetworkName);
                _localSubStorage.Set(KeySourceCampaign, value.CampaignName);
                _localSubStorage.Set(KeySourceGroup, value.GroupName);
            }
        }

        public DateTime NextDailyResetTime
        {
            get => _storage.Get(KeyNextDailyResetTime, default(DateTime));
            set => _storage.Set(KeyNextDailyResetTime, value);
        }

        public static GeneralPrefs Create([NotNull] string storagePath, [NotNull] string localSubStoragePath, [NotNull] PlatformAPI platform)
        {
            var storage = configure(BinaryStorage.Construct(storagePath)).Build();
            var localSubStorage = configure(BinaryStorage.Construct(localSubStoragePath)).Build();

            var prefs = new GeneralPrefs(storage, localSubStorage, platform);
            return prefs;

            [NotNull]
            BinaryStorage.Builder configure([NotNull] BinaryStorage.Builder storage)
            {
                return storage
                    .AddPrimitiveTypes()
                    .SupportNullable<bool>()
                    .SupportSetsOf<string>()
                    .SupportDictionariesOf<string, long>()
                    .SupportDictionariesOf<string, int>()
                    .SupportDictionariesOf<string, DateTime>()
                    .SupportEnum<SubscriptionStatus>()
                    .SupportEnum<InAppStatus>()
                    .SupportEnum<AuthorizationStatus>()
                    .SupportEnum<ConfigScope>();
            }
        }

        void ILegacyMigrator.Migrate(MigrationData migrationData)
        {
            using (MultipleChangeScope())
            {
                MigratedUser = true;
                ClientId.Value = migrationData.ClientId;
                FirstInstalledVersion = migrationData.FirstInstalledVersion;
                FirstLaunchDate = migrationData.FirstLaunchDate ?? default;
                SubscriptionStatus.Value = migrationData.SubscriptionStatus ?? default;
                InAppStatus.Value = migrationData.InAppStatus ?? default;
                AuthorizationStatus.Value = migrationData.AuthorizationStatus ?? default;
                ReferrerId.Value = migrationData.ReferrerId;
                AdjustId.Value = migrationData.AdjustId;
                GpsAdId.Value = migrationData.GpsAdId;
                IsGdprGranted.Value = migrationData.IsGdprGranted ?? default;
                AttAuthorizationStatus.Value = migrationData.AttAuthorizationStatus ?? default;
                UsedDefaultCampaignTypes.AddRange(migrationData.UsedDefaultCampaignTypes);
                ProductsReceiptTimestamps.AddRange(migrationData.ProductsReceiptTimestamps);
                PurchasedInAppProducts.AddRange(migrationData.PurchasedInAppProducts);
                PurchasedSubscriptionProducts.AddRange(migrationData.PurchasedSubscriptionProducts);
                UsedFreeProducts.AddRange(migrationData.UsedFreeProducts);

                FirebaseInstanceId.Value = migrationData.FirebaseInstanceId;
                GlobalSessionCounter.Value = migrationData.GlobalSessionCounter ?? default;
                SessionId.Value = migrationData.SessionId;
                SessionStarted = migrationData.SessionStarted ?? default;
                LastActivity = migrationData.LastActivity ?? default;
                MediaSource = migrationData.MediaSource ?? default;

                // Migrate from syncable to local (added during sync progress implementation)
                tryMigrateToLocal(KeyClientId, ClientId);
                tryMigrateToLocal(KeyAuthTokenExpired, AuthorizationTokenExpired);
                tryMigrateToLocal(KeyAuthToken, AuthorizationToken);
                tryMigrateToLocal(KeyInAppStatus, InAppStatus);
                tryMigrateToLocal(KeySubscriptionState, SubscriptionStatus);
                tryMigrateToLocal(KeyAuthorizationStatus, AuthorizationStatus);
                tryMigrateToLocal(KeyReferrerId, ReferrerId);
                tryMigrateToLocal(KeyLanguageCode, LanguageCode);
                tryMigrateToLocal(KeyCountryCode, CountryCode);
                tryMigrateToLocal(KeyGpsAdId, GpsAdId);
                tryMigrateToLocal(KeyGdprStatus, IsGdprGranted);
                tryMigrateToLocal(KeyAttAuthorizationStatus, AttAuthorizationStatus);
                tryMigrateToLocal(KeyAdjustId, AdjustId);
                tryMigrateToLocal(KeyFirebaseInstanceID, FirebaseInstanceId);

                void tryMigrateToLocal<T>([NotNull] string key, [NotNull] IReactiveProperty<T> property)
                {
                    if (_storage.Has<T>(key))
                    {
                        property.Value = _storage.Get<T>(key);
                        _storage.Remove<T>(key);
                    }
                }
            }
        }

        void IInitializable.Initialize()
        {
            if (string.IsNullOrEmpty(FirstInstalledVersion))
            {
                FirstInstalledVersion = Application.version;
            }
            if (FirstLaunchDate == default)
            {
                FirstLaunchDate = DateTime.Now;
            }
        }

        void IDisposable.Dispose()
        {
            _storage.Dispose();
            _localSubStorage.Dispose();
        }

        private GeneralPrefs([NotNull] BinaryStorage storage, [NotNull] BinaryStorage localSubStorage, [NotNull] PlatformAPI platform)
        {
            _storage = storage;
            _localSubStorage = localSubStorage;
            _platform = platform;
            using (MultipleChangeScope())
            {
                // General data, that must be synced with server
                GlobalSessionCounter = _storage.GetReactiveProperty<int>(KeySessionCounter, 0);
                SessionId = _storage.GetReactiveProperty<string>(KeySessionId, string.Empty);

                UsedDefaultCampaignTypes = _storage.GetSet<string>(KeyDefaultCampaignTypes);
                ProductsReceiptTimestamps = _storage.GetDictionary<string, DateTime>(KeyProductsReceiptTime);

                PurchasedInAppProducts = _storage.GetSet<string>(KeyPurchasedInApps);
                PurchasedSubscriptionProducts = _storage.GetSet<string>(KeyPurchasedSubscriptions);
                UsedFreeProducts = _storage.GetSet<string>(KeyUsedFreeProducts);
                AllUsedProducts = new UnionReadOnlyCollection<string>(PurchasedInAppProducts, PurchasedSubscriptionProducts, UsedFreeProducts);

                FirebaseInstanceId = _storage.GetReactiveProperty<string>(KeyFirebaseInstanceID);

                _versionSessionCounter = _storage.GetDictionary<string, int>(KeyVersionSessionCounter);
                _versionSessionCounter.TryAdd(Application.version, 0);

                // Local data, that won't be synced with server
                ClientId = _localSubStorage.GetReactiveProperty<string>(KeyClientId).AsNotNull(string.Empty);
                AuthorizationTokenExpired = _localSubStorage.GetReactiveProperty(KeyAuthTokenExpired, false);
                AuthorizationToken = _localSubStorage.GetReactiveProperty(KeyAuthToken, string.Empty);
                InAppStatus = _localSubStorage.GetReactiveProperty(KeyInAppStatus, DefaultInAppStatus);
                SubscriptionStatus = _localSubStorage.GetReactiveProperty(KeySubscriptionState, DefaultSubscriptionStatus);
                AuthorizationStatus = _localSubStorage.GetReactiveProperty(KeyAuthorizationStatus, DefaultAuthorizationStatus);
                ReferrerId = _localSubStorage.GetReactiveProperty(KeyReferrerId, string.Empty);
                LanguageCode = _localSubStorage.GetReactiveProperty(KeyLanguageCode, string.Empty);
                CountryCode = _localSubStorage.GetReactiveProperty(KeyCountryCode, string.Empty);
                GpsAdId = _localSubStorage.GetReactiveProperty<string>(KeyGpsAdId);
                IsGdprGranted = _localSubStorage.GetReactiveProperty<bool>(KeyGdprStatus);
                AttAuthorizationStatus = _localSubStorage.GetReactiveProperty<bool?>(KeyAttAuthorizationStatus);
                AdjustId = _localSubStorage.GetReactiveProperty(KeyAdjustId, default(string));
                FirebaseInstanceId = _localSubStorage.GetReactiveProperty<string>(KeyFirebaseInstanceID);
                SyncStateEnabled = _localSubStorage.GetReactiveProperty(KeySyncStateEnabled, false);
                RemoteContextScopes = _localSubStorage.GetReactiveProperty<ConfigScope>(KeyCampaignsContextScopes);
            }
        }

        /// <summary>
        /// If you are planning to change many properties of this class, you should use this method.
        /// This method temporarily blocks the writing prefs file to disk inside <b>using</b> construction.
        /// </summary>
        /// <example>
        /// <code>
        /// using (MultipleChangeScope())
        /// {
        ///     GeneralPrefs.ClientId = Guid.NewGuid().ToString();
        ///     GeneralPrefs.AuthorizationStatus.Value = Magify.AuthorizationStatus.Authorized;
        /// }
        /// </code>
        /// </example>
        /// <returns>IDisposable which must disposed automatically after <b>using</b> construction (preferred) or disposed manually.</returns>
        public IDisposable MultipleChangeScope()
        {
            return new DisposableTuple<IDisposable>((_storage.MultipleChangeScope(), _localSubStorage.MultipleChangeScope()));
        }

        public void Reset()
        {
            _storage.ResetAll();
            _localSubStorage.ResetAll();
        }

        /// <inheritdoc cref="BinaryStorage.GetFileContent"/>
        [NotNull]
        internal string GetFileContent()
        {
            return _storage.GetFileContent();
        }

        /// <inheritdoc cref="BinaryStorage.RewriteContent"/>
        internal void RewriteContent([NotNull] string contentBase64)
        {
            _storage.RewriteContent(contentBase64);
        }
    }
}